package com.cyclone.treecache;

import static org.hamcrest.MatcherAssert.assertThat;
import static org.hamcrest.Matchers.hasItem;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.concurrent.CountDownLatch;
import java.util.concurrent.Executor;
import java.util.concurrent.Executors;

import javax.transaction.TransactionManager;

import org.infinispan.Cache;
import org.infinispan.configuration.cache.CacheMode;
import org.infinispan.configuration.cache.Configuration;
import org.infinispan.configuration.cache.ConfigurationBuilder;
import org.infinispan.configuration.cache.VersioningScheme;
import org.infinispan.context.Flag;
import org.infinispan.manager.DefaultCacheManager;
import org.infinispan.transaction.LockingMode;
import org.infinispan.transaction.TransactionMode;
import org.infinispan.transaction.lookup.DummyTransactionManagerLookup;
import org.infinispan.tree.Node;
import org.infinispan.tree.TreeCache;
import org.infinispan.tree.TreeCacheFactory;
import org.junit.After;
import org.junit.Before;
import org.junit.Test;

import com.google.common.base.Throwables;

public class TreeCacheConcurrencyTests
{
    private static final String KEY = "someKey";

    private TransactionManager transactionManager;

    private TreeCache<String, String> treeCache;

    private DefaultCacheManager cacheManager;

    @Before
    public void createCache()
    {
        final Configuration config = createCacheConfig();

        cacheManager = new DefaultCacheManager(config);

        Cache<String, String> cache = cacheManager.getCache("cache");

        treeCache = new TreeCacheFactory().createTreeCache(cache);

        transactionManager = cache.getAdvancedCache().getTransactionManager();
    }

    private Configuration createCacheConfig()
    {
        return new ConfigurationBuilder().invocationBatching().enable()
                .clustering().cacheMode(CacheMode.LOCAL).transaction()
                .transactionManagerLookup(new DummyTransactionManagerLookup())
                .transactionMode(TransactionMode.TRANSACTIONAL)
                .useSynchronization(true).autoCommit(false)
                .lockingMode(LockingMode.OPTIMISTIC).versioning().enable()
                .scheme(VersioningScheme.SIMPLE).build();
    }

    @After
    public void destroyCache()
    {
        treeCache.stop();
        cacheManager.stop();
    }

    // These works...
    @Test
    public void serialPutsDoNotCorruptTree() throws Exception
    {
        putsDoNotCorruptTree(Executors.newSingleThreadExecutor());
    }

    @Test
    public void serialPutsDoNotCorruptTree_putForExternalRead()
            throws Exception
    {
        putsDoNotCorruptTree(Executors.newSingleThreadExecutor(),
                Flag.PUT_FOR_EXTERNAL_READ);
    }

    // ...these does not...
    @Test
    public void concurrentPutsDoNotCorruptTree() throws Exception
    {
        putsDoNotCorruptTree(Executors.newCachedThreadPool());
    }

    @Test
    public void concurrentPutsDoNotCorruptTree_putForExternalRead()
            throws Exception
    {
        putsDoNotCorruptTree(Executors.newCachedThreadPool(),
                Flag.PUT_FOR_EXTERNAL_READ);
    }

    private void putsDoNotCorruptTree(Executor executor, Flag... flags)
            throws Exception
    {
        final String parentFqn = "/parent";
        final int numChildren = 10;

        beginTx();
        treeCache.put(parentFqn, KEY, "parentValue", flags);
        commitTx();

        final CountDownLatch latch = new CountDownLatch(numChildren);

        for (String childFqn : childFqns(parentFqn, numChildren))
        {
            executor.execute(putTask(latch, childFqn));
        }

        latch.await();

        Collection<String> childFqnsFromActualNodes = actualChildFqns(treeCache
                .getNode(parentFqn));

        for (String childFqn : childFqns(parentFqn, numChildren))
        {
            // If the child data value is set successfully then the child node should be
            // present
            String value = treeCache.get(childFqn, KEY);
            if (value != null)
            {
                assertThat("Fqn with data should have a node that holds it",
                        childFqnsFromActualNodes, hasItem(childFqn));
            }
        }
    }

    private List<String> childFqns(final String parentFqn, final int numChildren)
    {
        List<String> childFqns = new ArrayList<>();
        for (int i = 0; i < numChildren; i++)
        {
            childFqns.add(parentFqn + "/child_" + i);
        }
        return childFqns;
    }

    private Collection<String> actualChildFqns(Node<String, String> parent)
    {
        Collection<String> fqns = new ArrayList<>();
        for (Node<String, String> child : parent.getChildren())
        {
            fqns.add(child.getFqn().toString());
        }

        return fqns;
    }

    private Runnable putTask(final CountDownLatch latch, final String fqn,
            final Flag... flags)
    {
        return new Runnable()
        {
            @Override
            public void run()
            {
                try
                {
                    beginTx();
                    treeCache.put(fqn, KEY, "value_" + fqn, flags);
                    commitTx();
                }
                finally
                {
                    latch.countDown();
                }
            }
        };
    }

    private final void beginTx()
    {
        try
        {
            transactionManager.begin();
        }
        catch (Exception e)
        {
            throw Throwables.propagate(e);
        }
    }

    private void commitTx()
    {
        try
        {
            transactionManager.commit();
        }
        catch (Exception e)
        {
            throw Throwables.propagate(e);
        }
    }
}
